@@ -138,8 +138,12 @@ tree_open_binpkg(const char *sroot, const char *spkg)
 		ret->cachetype = CACHE_BINPKGS;
 
 		snprintf(buf, sizeof(buf), "%s%s/%s", sroot, spkg, binpkg_packages);
-		if (eat_file(buf, &ret->pkgs, &ret->pkgslen))
+		if (eat_file(buf, &ret->pkgs, &ret->pkgslen)) {
 			ret->cachetype = CACHE_PACKAGES;
+		} else if (ret->pkgs != NULL) {
+			free(ret->pkgs);
+			ret->pkgs = NULL;
+		}
 	}
 
 	return ret;
@@ -148,6 +152,24 @@ tree_open_binpkg(const char *sroot, const char *spkg)
 void
 tree_close(tree_ctx *ctx)
 {
+	if (ctx->cache.categories != NULL) {
+		DECLARE_ARRAY(t);
+		size_t n;
+		tree_cat_ctx *cat;
+
+		values_set(ctx->cache.categories, t);
+		free_set(ctx->cache.categories);
+		ctx->cache.categories = NULL;  /* must happen before close_cat */
+
+		array_for_each(t, n, cat) {
+			/* ensure we cleanup all pkgs */
+			cat->pkg_cur = 0;
+			tree_close_cat(cat);
+		}
+
+		xarrayfree_int(t);
+	}
+
 	closedir(ctx->dir);
 	/* closedir() above does this for us: */
 	/* close(ctx->tree_fd); */
@@ -212,6 +234,21 @@ tree_open_cat(tree_ctx *ctx, const char *name)
 	int fd;
 	DIR *dir;
 
+	/* lookup in the cache, if any */
+	if (ctx->cache.categories != NULL) {
+		cat_ctx = get_set(name, ctx->cache.categories);
+		if (cat_ctx != NULL) {
+			/* reset state so it can be re-iterated (sort benefits the
+			 * most here) */
+			if (ctx->do_sort) {
+				cat_ctx->pkg_cur = 0;
+			} else {
+				rewinddir(cat_ctx->dir);
+			}
+			return cat_ctx;
+		}
+	}
+
 	/* Cannot use O_PATH as we want to use fdopendir() */
 	fd = openat(ctx->tree_fd, name, O_RDONLY | O_CLOEXEC);
 	if (fd == -1)
@@ -229,6 +266,15 @@ tree_open_cat(tree_ctx *ctx, const char *name)
 	cat_ctx->dir = dir;
 	cat_ctx->ctx = ctx;
 	cat_ctx->pkg_ctxs = NULL;
+	cat_ctx->pkg_cur = 0;
+	cat_ctx->pkg_cnt = 0;
+
+	if (ctx->cache.categories != NULL) {
+		add_set_value(name, cat_ctx, ctx->cache.categories);
+		/* ensure name doesn't expire after this instantiation is closed */
+		cat_ctx->name = contains_set(name, ctx->cache.categories);
+	}
+
 	return cat_ctx;
 }
 
@@ -289,6 +335,14 @@ tree_next_cat(tree_ctx *ctx)
 void
 tree_close_cat(tree_cat_ctx *cat_ctx)
 {
+	if (cat_ctx->ctx->cache.categories != NULL &&
+			contains_set(cat_ctx->name, cat_ctx->ctx->cache.categories))
+		return;
+
+	/* cleanup unreturned pkgs when sorted (or cache in use) */
+	while (cat_ctx->pkg_cur < cat_ctx->pkg_cnt)
+		tree_close_pkg(cat_ctx->pkg_ctxs[cat_ctx->pkg_cur++]);
+
 	closedir(cat_ctx->dir);
 	/* closedir() above does this for us: */
 	/* close(ctx->fd); */
@@ -403,7 +457,7 @@ tree_next_pkg_int(tree_cat_ctx *cat_ctx)
 				pkg_ctx = cat_ctx->pkg_ctxs[cat_ctx->pkg_cnt++] =
 					tree_open_pkg(cat_ctx, name);
 				if (pkg_ctx == NULL) {
-					free(name);
+					/* name was freed by tree_close_pkg on fail */
 					cat_ctx->pkg_cnt--;
 				}
 			}
@@ -499,7 +553,10 @@ tree_next_pkg(tree_cat_ctx *cat_ctx)
 		} while (ret == NULL);
 	} else if (ctx->cachetype == CACHE_BINPKGS) {
 		char *p = NULL;
+		ret = NULL;
 		do {
+			if (ret != NULL)
+				tree_close_pkg(ret);
 			ret = tree_next_pkg_int(cat_ctx);
 		} while (ret != NULL && (p = strstr(ret->name, ".tbz2")) == NULL);
 		if (p != NULL)
@@ -564,10 +621,11 @@ tree_read_file_pms(tree_pkg_ctx *pkg_ctx)
 		goto err;
 
 	len = sizeof(*ret) + s.st_size + 1;
-	ret = xzalloc(len);
+	ret = xmalloc(len);
 	ptr = (char*)ret + sizeof(*ret);
 	if ((off_t)fread(ptr, 1, s.st_size, f) != s.st_size)
 		goto err;
+	ptr[s.st_size] = '\0';
 
 	ret->Q_DEPEND = ptr;
 #define next_line(curr, next) \
@@ -631,10 +689,11 @@ tree_read_file_md5(tree_pkg_ctx *pkg_ctx)
 		goto err;
 
 	len = sizeof(*ret) + s.st_size + 1;
-	ret = xzalloc(len);
+	ret = xmalloc(len);
 	ptr = (char*)ret + sizeof(*ret);
 	if ((off_t)fread(ptr, 1, s.st_size, f) != s.st_size)
 		goto err;
+	ptr[s.st_size] = '\0';
 
 	/* We have a block of key=value\n data.
 	 * KEY=VALUE\n
@@ -743,10 +802,11 @@ tree_read_file_ebuild(tree_pkg_ctx *pkg_ctx)
 		goto err;
 
 	len = sizeof(*ret) + s.st_size + 1;
-	ret = xzalloc(len);
+	ret = xmalloc(len);
 	p = (char *)ret + sizeof(*ret);
 	if ((off_t)fread(p, 1, s.st_size, f) != s.st_size)
 		goto err;
+	p[s.st_size] = '\0';
 
 	do {
 		q = p;
@@ -1108,13 +1168,14 @@ tree_pkg_metadata(tree_pkg_ctx *pkg_ctx)
 	}
 
 	len = sizeof(*ret) + s.st_size + 1;
-	p = xbuf = xzalloc(len);
+	p = xbuf = xmalloc(len);
 	if ((off_t)fread(p, 1, s.st_size, f) != s.st_size) {
 		free(p);
 		fclose(f);
 		pkg_ctx->fd = -1;
 		return NULL;
 	}
+	p[s.st_size] = '\0';
 
 	ret = xmalloc(sizeof(*ret));
 	ret->email = NULL;
@@ -1186,10 +1247,12 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 
 	/* reused for every entry */
 	tree_cat_ctx *cat = NULL;
-	tree_pkg_ctx *pkg = xzalloc(sizeof(tree_pkg_ctx));
-	tree_pkg_meta *meta = xzalloc(sizeof(tree_pkg_meta));
+	tree_pkg_ctx pkg;
+	tree_pkg_meta meta;
 	depend_atom *atom = NULL;
 
+	memset(&meta, 0, sizeof(meta));
+
 	do {
 		/* find next line */
 		c = NULL;
@@ -1206,32 +1269,41 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 			if (atom != NULL) {
 				size_t pkgnamelen;
 
+				memset(&pkg, 0, sizeof(pkg));
+
 				/* store meta ptr in repo->pkgs, such that get_pkg_meta
 				 * can grab it from there (for free) */
-				ctx->pkgs = (char *)meta;
+				c = ctx->pkgs;
+				ctx->pkgs = (char *)&meta;
 
 				if (cat == NULL || strcmp(cat->name, atom->CATEGORY) != 0)
 				{
-					if (cat != NULL)
+					if (cat != NULL) {
+						atom_implode((depend_atom *)cat->pkg_ctxs);
+						cat->pkg_ctxs = NULL;
 						tree_close_cat(cat);
-					pkg->cat_ctx = cat = tree_open_cat(ctx, atom->CATEGORY);
+					}
+					pkg.cat_ctx = cat = tree_open_cat(ctx, atom->CATEGORY);
+					cat->pkg_ctxs = (tree_pkg_ctx **)atom;  /* for name */
 				}
 				pkgnamelen = snprintf(pkgname, sizeof(pkgname),
 						"%s.tbz2", atom->PF);
 				pkgname[pkgnamelen - (sizeof(".tbz2") - 1)] = '\0';
-				pkg->name = pkgname;
-				pkg->slot = meta->Q_SLOT == NULL ? (char *)"0" : meta->Q_SLOT;
-				pkg->repo = ctx->repo;
-				pkg->atom = atom;
-				pkg->fd = 0;  /* intentional, meta has already been read */
+				pkg.name = pkgname;
+				pkg.slot = meta.Q_SLOT == NULL ? (char *)"0" : meta.Q_SLOT;
+				pkg.repo = ctx->repo;
+				pkg.atom = atom;
+				pkg.fd = 0;  /* intentional, meta has already been read */
 
 				/* do call callback with pkg_atom (populate cat and pkg) */
-				ret |= callback(pkg, priv);
+				ret |= callback(&pkg, priv);
 
-				atom_implode(atom);
+				ctx->pkgs = c;
+				if (atom != (depend_atom *)cat->pkg_ctxs)
+					atom_implode(atom);
 			}
 
-			memset(meta, 0, sizeof(meta[0]));
+			memset(&meta, 0, sizeof(meta));
 			atom = NULL;
 			if (len > 0) {  /* hop over \n */
 				p++;
@@ -1267,7 +1339,7 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 #define match_key(X) match_key2(X,X)
 #define match_key2(X,Y) \
 		} else if (strcmp(p, #X) == 0) { \
-			meta->Q_##Y = c
+			meta.Q_##Y = c
 		match_key(DEFINED_PHASES);
 		match_key(DEPEND);
 		match_key2(DESC, DESCRIPTION);
@@ -1288,8 +1360,18 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 		p = q;
 	} while (len > 0);
 
+	if (cat != NULL) {
+		atom_implode((depend_atom *)cat->pkg_ctxs);
+		cat->pkg_ctxs = NULL;
+		tree_close_cat(cat);
+	}
+
+	if (atom != NULL)
+		atom_implode(atom);
+
 	/* ensure we don't free a garbage pointer */
 	ctx->repo = NULL;
+	ctx->do_sort = false;
 
 	return ret;
 }
@@ -1439,3 +1521,35 @@ tree_get_atoms(tree_ctx *ctx, bool fullcpv, set *satoms)
 
 	return state.cpf;
 }
+
+tree_pkg_ctx *
+tree_match_atom(tree_ctx *ctx, depend_atom *a)
+{
+	tree_cat_ctx *cat_ctx;
+	tree_pkg_ctx *pkg_ctx;
+	depend_atom *atom;
+
+	if (ctx->cache.categories == NULL)
+		ctx->cache.categories = create_set();
+
+	if (a->P == NULL) {
+		return NULL;
+	} else if (a->CATEGORY == NULL) {
+		/* loop through all cats and recurse */
+		/* TODO: some day */
+		return NULL;
+	} else {
+		/* try CAT, and PN for latest version */
+		if ((cat_ctx = tree_open_cat(ctx, a->CATEGORY)) == NULL)
+			return NULL;
+		ctx->do_sort = true;     /* sort uses buffer, which cache relies on */
+		ctx->query_atom = NULL;  /* ensure the cache contains ALL pkgs */
+		while ((pkg_ctx = tree_next_pkg(cat_ctx)) != NULL) {
+			atom = tree_get_atom(pkg_ctx, a->SLOT != NULL);
+			if (atom_compare(atom, a) == EQUAL)
+				return pkg_ctx;
+		}
+
+		return NULL;
+	}
+}
